@@ -4,7 +4,7 @@
  *   Copyright (C) 1996 Roger Espel Llima <roger.espel.llima@pobox.com>
  *
  *   Started: 10 Oct 1996
- *   Version: 1.01
+ *   Version: 1.14
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -12,7 +12,7 @@
  *
  */
 
-#define VERSION "1.01"
+#define VERSION "1.14"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -37,7 +37,7 @@
 
 #define PROT_VERS_1 1
 #define PROT_VERS_2 0
-#define PROT_VERS_3 0
+#define PROT_VERS_3 1
 
 #define OP_NICK 1
 #define OP_PLAY 2
@@ -60,6 +60,8 @@
 #define OP_MSG 19
 #define OP_YOUARE 20
 #define OP_LINESTO 21
+#define OP_WON 22
+#define OP_ZERO 23
 
 
 static int shape[7][4][8] = {
@@ -136,6 +138,8 @@ int hscroll = 10;
 int paused = 0;
 int visiblemsgs = 0, msgsclosed = 0;
 int verbose = 0;
+int norestart = 0;
+int nospeedup = 0;
 
 /* X stuff */
 char *disp = NULL;
@@ -151,7 +155,7 @@ int keysspecified = 0;
 KeySym quit_key = XK_q, start_key = XK_s, pause_key = XK_p, bot_key = XK_b;
 KeySym text_key = XK_t, togglenext_key = XK_Return, mode_key = XK_n;
 KeySym verbose_key = XK_v, clear_key = XK_c, warp_key = XK_Tab;
-KeySym down_key = XK_comma;
+KeySym down_key = XK_comma, zero_key = XK_z;
 KeySym level_keys[10] = 
   { XK_0, XK_1, XK_2, XK_3, XK_4, XK_5, XK_6, XK_7, XK_8, XK_9 };
 KeySym keys[6] = { XK_J, XK_j, XK_k, XK_l, XK_L, XK_space };
@@ -173,9 +177,14 @@ int warp_next = 1;
 #define XNEXT 15
 #define XNEXTTXT 16
 #define XNICK (XMARGIN + (WPITIN-WNICK) / 2) /* relative to u->x0 */
-#define XMSGS 30
+#define XTXT 30
+#define XZERO 30
 #define XSCROLL (XPIT + LEFTSIDE)
 #define XMSGENTRY (XPIT + LEFTSIDE)
+#define XLINESTXT (XMARGIN + 25)
+#define XLINES (XMARGIN + 80)
+#define XGAMES XLINES
+#define XGAMESTXT (XMARGIN + 25)
 
 #define YQUIT 13
 #define YPLAY 40
@@ -187,7 +196,12 @@ int warp_next = 1;
 #define YNEXTTXT 279
 #define YNEXT 285
 #define YNICK (16 + YMARGIN + (HSQR * ROWS))
-#define YMSGS YNICK
+#define YTXT YNICK
+#define YZERO (YTXT + 35)
+#define YLINES (YNICK + HTEXTENTRIES + 13)
+#define YLINESTXT (YLINES + 15)
+#define YGAMES (YLINES + HTEXTENTRIES + 10)
+#define YGAMESTXT (YGAMES + 15)
 #define YSCROLL (HEIGHT0 + 10)
 #define YMSGENTRY (YSCROLL + HSCROLL * FONTHEIGHT + 17)
 #define YVERBOSE (YSCROLL + 20)
@@ -201,14 +215,15 @@ int warp_next = 1;
 #define LEFTSIDE (46 + 4 * WSQR) /* size of the whole left-side buttons area */
 
 #define WIDTH0 ((WSQR * COLS) + XMARGIN + 22)
-#define HEIGHT0 (52 + YMARGIN + (HSQR * ROWS))
+#define HEIGHT0 (117 + YMARGIN + (HSQR * ROWS))
 
 #define WSQR sqrsize
 #define HSQR sqrsize
 
 #define WBTNS 70
 #define HBTNS 20
-#define WMSGS 50
+#define WTXT 50
+#define WZERO 50
 #define WNICKCHARS 16
 
 #define WNICK (WNICKCHARS * FONTWIDTH + 8)
@@ -221,6 +236,10 @@ int warp_next = 1;
 #define WSCROLLRESIZE (X_EACH / FONTWIDTH)
 #define HSCROLLRESIZE (HSCROLL * FONTHEIGHT + HTEXTENTRIES + 45)
 #define WMSGENTRY (WSCROLL * FONTWIDTH + 8)
+#define WLINES 60
+#define HLINES HTEXTENTRIES
+#define WGAMES WLINES
+#define HGAMES HTEXTENTRIES
 #define HTEXTENTRIES 23
 
 /* don't change these: */
@@ -318,16 +337,17 @@ struct flashlinedata {
 
 /* server stuff */
 
-#define BUFLEN 1024
+#define BUFLEN 4096
 int sfd;
 int port = 0;
-unsigned char realbuf[512], readbuf[BUFLEN], *buf = realbuf + 4;
-int nread = 0;
+unsigned char realbuf[512], readbuf[BUFLEN], writbuf[BUFLEN];
+unsigned char *buf = realbuf + 4;
+int nread = 0, nwrite = 0;
 char *serverhost = NULL;
 
 struct textentry *nickentry, *msgentry;
 struct button *modebutton, *pausebutton, *quitbutton, *playbutton, *botbutton;
-struct button *textbutton = NULL, *verbosebutton, *clearbutton;
+struct button *textbutton = NULL, *verbosebutton, *clearbutton, *zerobutton;
 struct toggle *leveltogs[10];
 struct textscroll *msgscroll;
 int scrollcols; /* for 1 user */
@@ -341,6 +361,7 @@ struct user {
   char haslost, flashing;
   char replaybuf[BUFLEN];
   int replaylen;
+  int lines, games;
 };
 
 struct user me;
@@ -388,12 +409,22 @@ void u_sleep(int i) {
   select(0, NULL, NULL, NULL, &tm);
 }
 
+void flushbuf() {
+  if (nwrite) {
+    write(sfd, writbuf, nwrite);
+    nwrite = 0;
+  }
+}
+
 void sendbuf(int len) {
   if (!standalone) {
     realbuf[0] = realbuf[1] = realbuf[2] = 0;
     realbuf[3] = (unsigned char)len;
     buf[0] = 0;
-    write(sfd, realbuf, 4 + len);
+    if (nwrite + 4 + len >= BUFLEN)
+      fatal("Internal error: send buffer overflow");
+    memcpy(writbuf + nwrite, realbuf, 4 + len);
+    nwrite += 4 + len;
   }
 }
 
@@ -432,11 +463,14 @@ void getnumresource(char *r, int *i) {
 }
 
 void getresources() {
+  char *s;
+
   if (!keysspecified) {
     getkeyresource("quitKey", &quit_key);
     getkeyresource("startKey", &start_key);
     getkeyresource("pauseKey", &pause_key);
     getkeyresource("botKey", &bot_key);
+    getkeyresource("zeroKey", &zero_key);
     getkeyresource("textKey", &text_key);
     getkeyresource("togglenextKey", &togglenext_key);
     getkeyresource("clearKey", &clear_key);
@@ -468,6 +502,10 @@ void getresources() {
   getnumresource("scrollLines", &hscroll);
   if (hscroll < 3 || hscroll > 30)
     hscroll = 10;
+
+  s = XGetDefault(display, "xtris", "name");
+  if (s)
+    strncpy(me.nick, s, MAXNICKLEN);
 }
 
 int GetColor(char *s) {
@@ -758,10 +796,10 @@ void DrawButton(int updown, int bar, int x, int y, int w, int h, char *text, int
 	xx = x + ((w - textwidth(text)) >> 1);
 	break;
       default:
-	xx = x + w - textwidth(text) - 4;
+	xx = x + w - textwidth(text) - 6;
 	break;
     }
-    XDrawString(display, win, gc, xx, y+FONTHEIGHT+1 , text, strlen(text));
+    XDrawString(display, win, gc, xx, y+FONTHEIGHT+2 , text, strlen(text));
   }
 }
 
@@ -1121,7 +1159,7 @@ void ChangeTextEntryArea(struct textentry *t, char *txt, int align, int cursor)
 void FlashButton(struct button *b) {
   DrawButton(!b->down, 0, b->x, b->y, b->w, b->h, b->txt, b->align, 0);
   XFlush(display);
-  u_sleep(120000);
+  u_sleep(100000);
   DrawButton(b->down, 0, b->x, b->y, b->w, b->h, b->txt, b->align, 0);
   XFlush(display);
 }
@@ -1138,6 +1176,20 @@ void SimulateTextEntry(struct textentry *t, char *txt) {
   t->next_clears = 1;
 }
 
+void updatelines(struct user *u) {
+  static char tmpbuf[12];
+  
+  sprintf(tmpbuf, "%d", u->lines);
+  DrawButton(1, 1, XLINES + u->x0, YLINES, WLINES, HLINES, tmpbuf, ALIGN_RIGHT, 1);
+}
+
+void updategames(struct user *u) {
+  static char tmpbuf[12];
+  
+  sprintf(tmpbuf, "%d", u->games);
+  DrawButton(1, 1, XGAMES + u->x0, YGAMES, WGAMES, HGAMES, tmpbuf, ALIGN_RIGHT, 1);
+}
+
 void DrawFrame() {
   struct user *u;
 
@@ -1157,6 +1209,11 @@ void DrawFrame() {
     DrawButton(1, 0, XMARGIN + u->x0, YMARGIN, WPITOUT, HPITOUT, NULL, ALIGN_LEFT, 0);
     if (u != &me)  /* me.nick is actually a textarea */
       DrawButton(1, 0, XNICK + u->x0, YNICK, WNICK, HTEXTENTRIES, u->nick, ALIGN_CENTER, 0);
+    SetColor(color[TEXT][0]);
+    WriteXY(XLINESTXT + u->x0, YLINESTXT, "Lines:");
+    WriteXY(XGAMESTXT + u->x0, YGAMESTXT, "Games:");
+    updatelines(u);
+    updategames(u);
   }
 }
 
@@ -1194,6 +1251,15 @@ int fits(int piece, int rotation, int x, int y) {
   return 1;
 }
 
+int sticksout(int piece, int rotation, int x, int y) {
+  int i;
+
+  for (i=0; i<4; i++)
+    if (y - shape[piece][rotation][2*i+1] < 0)
+      return 1;
+  return 0;
+}
+
 void put(int piece, int rotation, int x, int y, int color) {
   int xx, yy, i;
 
@@ -1250,11 +1316,37 @@ void refresh() {
     sendbuf(i);
 }
 
-void refreshme(int n) {
+void refreshnet() {
+  int x, y;
+  int i = 2;
+
+  buf[1] = OP_DRAW;
+
+  for (x=0; x<10; x++)
+    for (y=0; y<20; y++)
+      if (me.pit[y][x] != oldpit[y][x]) {
+	if (me.pit[y][x] != (oldpit[y][x] & 0xff)) {
+	  buf[i] = x;
+	  buf[i+1] = y;
+	  buf[i+2] = me.pit[y][x];
+	  i += 3;
+	  if (i > 200) {
+	    sendbuf(i);
+	    i = 2;
+	    buf[1] = OP_DRAW;
+	  }
+	}
+	oldpit[y][x] = me.pit[y][x];
+      }
+  if (i > 2)
+    sendbuf(i);
+}
+
+void refreshme() {
   int x, y;
 
   for (x=0; x<10; x++)
-    for (y=0; y<n; y++)
+    for (y=0; y<20; y++)
       if (me.pit[y][x] != oldpit[y][x]) {
 	DrawSqr(&me, me.pit[y][x], x, y);
 	oldpit[y][x] = me.pit[y][x];
@@ -1332,10 +1424,12 @@ void fallines() {
       for (i=0; i<10; i++)
 	oldpit[n][i] |= 0x100;
       buf[2 + fallen++] = (unsigned char)n;
+      me.lines++;
     }
   }
   if (fallen) {
     sendbuf(2 + fallen);
+    flushbuf();
     XFlush(display);
     u_sleep(80000);
     nextcol = (nextcol+1) % 7;
@@ -1348,10 +1442,11 @@ void fallines() {
       DrawLine(&me, nextcol, buf[i]);
     u_sleep(80000);
     XFlush(display);
-    refreshme(20);
+    refreshme();
+    updatelines(&me);
     XFlush(display);
   }
-  if (fallen > 1 || (fallen > 0 && funmode)) {
+  if (fallen > 1 || (funmode && fallen > 0)) {
     buf[1] = OP_LINES;
     buf[2] = (unsigned char)fallen;
     sendbuf(3);
@@ -1380,8 +1475,11 @@ int newlines() {
       i--;
     }
     if (addlines > 0) {
-      refreshme(20 - addlines);
-      refresh();
+      refreshme();
+      for (j=20-addlines; j<20; j++)
+	for (k=0; k<10; k++)
+	  oldpit[j][k] = EMPTY;
+      refreshnet();
       XFlush(display);
     }
     addlines = 0;
@@ -1466,7 +1564,7 @@ void c_msgentry(char *s, int shift) {
   }
 }
 
-void c_msgs() {
+void c_txt() {
   if (visiblemsgs) {
     height -= HSCROLLRESIZE;
     visiblemsgs = 0;
@@ -1479,6 +1577,17 @@ void c_msgs() {
   redrawall();
 }
 
+void c_zero() {
+  if (standalone) {
+    me.games = 0;
+    updategames(&me);
+    XFlush(display);
+  } else {
+    buf[1] = OP_ZERO;
+    sendbuf(2);
+  }
+}
+
 void c_quit() {
   exit(0);
 }
@@ -1557,15 +1666,15 @@ void c_setnick(char *s, int shift) {
 }
 
 void setlevel(int n) {
-  if (n < 0)
-    n = 0;
+  if (n < 1)
+    n = 1;
   if (n > 9)
     n = 9;
   level = n;
   ToggleDown(leveltogs[n]);
 }
 
-void setmode(int n) {
+void set_mode(int n) {
   funmode = (n != 0);
   if (funmode)
     ChangeButtonAppearance(modebutton, -1, "fuN", -1);
@@ -1585,7 +1694,7 @@ void sendlevel(int n) {
 
 void sendmode(int n) {
   if (standalone)
-    setmode(n);
+    set_mode(n);
   else {
     buf[1] = OP_MODE;
     buf[2] = n;
@@ -1709,7 +1818,7 @@ void textentrykey(struct textentry *t, int keysym, unsigned int mask) {
 	SetColor(color[EMPTY][1]);
 	XFillRectangle(display, win, gc, t->x + FONTWIDTH * (len - t->offset) + 4, t->y + 3, FONTWIDTH + 3, HTEXTENTRIES - 6);
 	SetColor(color[TEXT][2]);
-	XDrawString(display, win, gc, t->x + FONTWIDTH * (len - t->offset) + 4, t->y + FONTHEIGHT + 1, &t->txt[len], 1);
+	XDrawString(display, win, gc, t->x + FONTWIDTH * (len - t->offset) + 4, t->y + FONTHEIGHT + 2, &t->txt[len], 1);
 	XDrawLine(display, win, gc, t->x + FONTWIDTH * (len - t->offset + 1) + 6, t->y + 4, t->x + FONTWIDTH * (len - t->offset + 1) + 6, t->y + 16);
 	XFlush(display);
       } else
@@ -1768,6 +1877,17 @@ void doevent(XEvent *xev) {
 	  break;
 	}
       }
+      if (px >= XNEXT && px <= XNEXT + WNEXT && py >= YNEXT &&
+	  py <= YNEXT + HNEXT && playing) {
+	if (shownext) {
+	  clearnext();
+	  shownext = 0;
+	} else {
+	  shownext = 1;
+	  drawnext();
+	}
+      }
+
       break;
 
     case KeyPress:
@@ -1834,6 +1954,11 @@ void doevent(XEvent *xev) {
 	break;
       }
 
+      if (keysym == zero_key) {
+	SimulateButtonPress(zerobutton);
+	break;
+      }
+
       if (keysym == text_key && textbutton != NULL) {
 	SimulateButtonPress(textbutton);
 	break;
@@ -2023,6 +2148,8 @@ void waitfor(int delay) {
   }
 
   do {
+    flushbuf();
+
     gettimeofday(&now, NULL);
 
     if (timer0) {
@@ -2229,6 +2356,7 @@ void handlemessages() {
 	}
 	v->next = u = mmalloc(sizeof(struct user));
 	u->number = buf[0];
+	u->games = (len >= 4 ? buf[2] * 256 + buf[3] : 0);
 	u->nick[0] = 0;
 	u->next = NULL;
 	for (i=0; i<10; i++)
@@ -2239,7 +2367,7 @@ void handlemessages() {
 	width += X_EACH;
 
 	if (!textbutton)
-	  textbutton = DeclareButton(XMSGS, YMSGS, WMSGS, HBTNS, 0, 1, "Text", ALIGN_CENTER, c_msgs);
+	  textbutton = DeclareButton(XTXT, YTXT, WTXT, HBTNS, 0, 1, "Text", ALIGN_CENTER, c_txt);
 	
 	scrollcols += WSCROLLRESIZE;
 	SetTextScrollWidth(msgscroll, scrollcols);
@@ -2378,6 +2506,25 @@ void handlemessages() {
 	sysmsg(msgbuf);
 	break;
 
+      case OP_WON:
+	u = finduser(buf[0]);
+	sprintf(msgbuf, "%s %s the game", getid(buf[0]),
+			isyou ? "win" : "wins");
+	sysmsg(msgbuf);
+	u->games++;
+	updategames(u);
+	break;
+
+      case OP_ZERO:
+	sprintf(msgbuf, "%s reset%s game counters", getid(buf[0]), 
+		isyou ? "" : "s");
+	sysmsg(msgbuf);
+	for (v=&me; v; v=v->next) {
+	  v->games = 0;
+	  updategames(v);
+	}
+	break;
+
       case OP_NICK:
 	u = finduser(buf[0]);
 	buf[len] = 0;
@@ -2424,6 +2571,8 @@ void handlemessages() {
 	  fld->nlines = len - 2;
 	  fld->color = i = my_rand() % 7;
 	  u->flashing = 1;
+	  u->lines += len - 2;
+	  updatelines(u);
 	  for (j=2; j<len; j++)
 	    DrawLine(u, i, buf[j]);
 	  XFlush(display);
@@ -2482,7 +2631,7 @@ void handlemessages() {
 	s = getid(buf[0]);
 	sprintf(msgbuf, "%s set%s %s mode", s, isyou ? "" : "s", buf[2] ? "fun" : "normal");
 	sysmsg(msgbuf);
-	setmode(buf[2]);
+	set_mode(buf[2]);
 	break;
     }
   }
@@ -2502,20 +2651,32 @@ void doserverstuff() {
 }
 
 void startserver() {
+  char *options[2];
+  int n = 0;
+
+  options[0] = options[1] = NULL;
+  if (norestart)
+    options[n++] = "-norestart";
+  if (nospeedup)
+    options[n++] = "-nospeedup";
+
   switch (fork()) {
     case 0:
+      execlp(
 #ifdef XTRISPATH
-      execl(XTRISPATH "/xtserv", "xtserv", "-once", NULL);
+      XTRISPATH "/xtserv",
 #else
-      execlp("xtserv", "xtserv", "-once", NULL);
+      "xtserv",
 #endif
-      fprintf(stderr, "Can't start server '%s'.\n",
+      "xtserv", "-once", options[0], options[1], NULL);
 
+      fprintf(stderr, "Can't start server '%s'.\n",
 #ifdef XTRISPATH
 	XTRISPATH "/xtserv");
 #else
 	"xtserv");
 #endif
+
       _exit(1);
     
     case -1:
@@ -2578,7 +2739,7 @@ void sigchld() {
 
 int main(int argc, char *argv[]) {
   int i, j;
-  char *s;
+  char *s, *setnick = NULL;
 
 #ifndef NeXT
   struct sigaction sact;
@@ -2612,10 +2773,13 @@ int main(int argc, char *argv[]) {
 "Use: xtris [options] [server.name [port]]\n"
 "Options:   -display machine:0   --  set the display\n"
 "           -k keys              --  set keys (default is \"JjklL \")\n"
+"           -n nick              --  sets the nickname of the player\n"
 "           -s                   --  standalone mode\n"
 "           -bw                  --  force black & white\n"
 "           -big                 --  bigger squares\n"
 "           -flashy              --  use flashier colors\n"
+"           -norestart           --  pass a -norestart to the server\n"
+"           -nospeedup           --  pass a -nospeedup to the server\n"
 
 );
 
@@ -2630,10 +2794,24 @@ int main(int argc, char *argv[]) {
       disp = argv[2];
       argv += 2;
       argc -= 2;
+    } else if (strcmp(argv[1], "-n") == 0) {
+      if (argc < 3)
+	fatal("Missing argument for -n");
+      setnick = argv[2];
+      argv += 2;
+      argc -= 2;
     } else if (strcmp(argv[1], "-s") == 0) {
       standalone = 1;
       argv++;
       argc--;
+    } else if (strcmp(argv[1], "-norestart") == 0) {
+      norestart = 1;
+      argv++;
+      argc--;
+    } else if (strcmp(argv[1], "-nospeedup") == 0) {
+      nospeedup = 1;
+      argv++;
+      argc--;
     } else if (strcmp(argv[1], "-k") == 0) {
       if (argc < 3)
 	fatal("Missing argument for -k");
@@ -2666,6 +2844,8 @@ int main(int argc, char *argv[]) {
   }
 
   me.number = 0;
+  me.lines = 0;
+  me.games = 0;
   me.x0 = LEFTSIDE;
 
   if ((s = getenv("XTRISNAME")) || (s = getenv("NETTRISNAME")) || 
@@ -2686,6 +2866,9 @@ int main(int argc, char *argv[]) {
 
   init_X();
 
+  if (setnick != NULL)
+    strncpy(me.nick, setnick, MAXNICKLEN);
+
   quitbutton = DeclareButton(XBTNS, YQUIT, WBTNS, HBTNS, 0, 1, "Quit", ALIGN_CENTER, c_quit);
   playbutton = DeclareButton(XBTNS, YPLAY, WBTNS, HBTNS, 0, 1, "Start", ALIGN_CENTER, c_play);
   pausebutton = DeclareButton(XBTNS, YPAUSE, WBTNS, HBTNS, 0, 0, "Pause", ALIGN_CENTER, c_pause);
@@ -2711,6 +2894,8 @@ int main(int argc, char *argv[]) {
 
   msgscroll = DeclareTextScroll(XSCROLL, YSCROLL, WSCROLL, HSCROLL, color[TEXT][2], msgentry);
 
+  zerobutton = DeclareButton(XZERO, YZERO, WZERO, HBTNS, 0, 1, "Zero", ALIGN_CENTER, c_zero);
+
   nickentry = DeclareTextEntry(XNICK + me.x0, YNICK, WNICKCHARS, MAXNICKLEN, "", 0, c_setnick);
   SimulateTextEntry(nickentry, me.nick);
 
@@ -2746,6 +2931,8 @@ int main(int argc, char *argv[]) {
       u->haslost = 0;
       u->flashing = 0;
       u->replaylen = 0;
+      u->lines = 0;
+      updatelines(u);
     }
     for (i=0; i<20; i++)
       for (j=0; j<10; j++)
@@ -2787,6 +2974,8 @@ int main(int argc, char *argv[]) {
 	y++;
       }
       put(piece, rotation, x, --y, piece);
+      if (sticksout(piece, rotation, x, y))
+	break;
       fallines();
       if (!newlines() || !newsquares())
 	break;
@@ -2794,6 +2983,10 @@ int main(int argc, char *argv[]) {
     buf[1] = OP_LOST;
     sendbuf(2);
     sysmsg("you have lost");
+    if (standalone) {
+      me.games++;
+      updategames(&me);
+    }
     clearnext();
     falldown(&me);
   }
